import os

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import LSTM, Dense, Dropout
import time
from series_extractor import TimeSeriesExtractor

class LSTMModel:
    def __init__(self, look_back=10, epochs=100, batch_size=32, output_lstm=True):
        self.look_back = look_back
        self.epochs = epochs
        self.batch_size = batch_size
        self.output_lstm = output_lstm
        os.makedirs(self.output_lstm, exist_ok=True)

    def create_dataset(self, data):
        X, Y = [], []
        for i in range(len(data) - self.look_back):
            X.append(data[i:(i + self.look_back), :])
            Y.append(data[i + self.look_back, 0])  # Assuming predicting Y
        return np.array(X), np.array(Y)

    def build_model(self, input_shape):
        model = Sequential()
        model.add(LSTM(units=50, input_shape=input_shape))
        model.add(Dense(units=32, activation='relu'))
        model.add(Dense(units=1, activation='linear'))
        model.add(Dropout(0.1))
        model.add(Dense(units=1))
        model.compile(optimizer='adam', loss='mean_squared_error')
        return model

    def compute_rmse(self, y_true, y_pred):
        mae = mean_absolute_error(y_true, y_pred)
        mse = mean_squared_error(y_true, y_pred)
        rmse = np.sqrt(mse)
        r2 = r2_score(y_true, y_pred)

        print("mean y_test:", np.mean(y_true))
        print("MSE:", mse)
        print("MAE:", mae)
        print("RMSE:", rmse)
        print("R2 score:", r2)

        rmse_percent = rmse / np.mean(y_true) * 100

        print("MSE %:", mse / np.mean(y_true) * 100)
        print("MAE %:", mae / np.mean(y_true) * 100)
        print("RMSE %:", rmse_percent)

        return rmse_percent

    def plot_training_history(self, history, n_var, col, normx, area):
        loss = history.history['loss']
        plt.figure(figsize=(10, 6))
        plt.plot(range(len(loss)), loss)
        plt.xlabel('Epochs')
        plt.ylabel('Loss: Mean Squared Error')
        plt.title(f'Loss_{n_var}_INTPP_to_{col}_{normx}_{area}')
        plt.savefig(f'{self.output_lstm}/LSTM_Loss_{n_var}_INTPP_to_{col}_{normx}_{area}.png')

        if 'accuracy' in history.history:
            acc = history.history['accuracy']
            plt.figure(figsize=(10, 6))
            plt.plot(range(len(acc)), acc)
            plt.xlabel('Epochs')
            plt.ylabel('Accuracy')
            plt.title(f'ACC_{n_var}_INTPP_to_{col}_{normx}_{area}')
            plt.savefig(f'{self.output_lstm}/LSTM_Acc_{n_var}_INTPP_to_{col}_{normx}_{area}.png')

    def plot_forecast(self, Y_test_transformed, y_pred_transformed, n_var, col, normx, area, rmse):

        plt.plot(Y_test_transformed.flatten(), label='Test sequence', linestyle='dotted')
        plt.plot(y_pred_transformed.flatten(), label='Predicted sequence')
        plt.title(f'ML LSTM Multvar. INTPP to {col.upper()} & {normx.upper()} @ {area} forecast. RMSE: {rmse:.2f} %')
        plt.xlabel('Time')
        plt.ylabel('Transformed values [ ]')
        plt.legend()
        plt.savefig(f'{self.output_lstm}/LSTM_forecast_var_{n_var}_INTPP_to_{col}_{normx}_{area}.png')

    def run_lstm(self, df_train, df_test, area_dict, normx_dict, columns, variables, n_forecast_steps):
        start_time_lstm = time.time()
        rmse_results = pd.DataFrame(columns=['NORM', 'COLUMNS', 'VAR', 'RMSE', 'AREA', 'MODEL', 'HYPERTUNE'])

        for area in area_dict:
            print('RUN FOR FOLLOWING LOCATION ', area)
            for normx in normx_dict:

                print("df_train.head() ",df_train.head())
                print("df_test.head() ",df_test.head())
                print("type(variables)", type(variables))
                print("columns ",columns)
                for col in range(0,2):
                    print("f'{variables[col].lower()}_{normx}_train_{area}'",f'{variables[col].lower()}_{normx}_train_{area}')
                    print("variables[col].lower() ",variables[col].lower())
                print("LOOP ", {col: f'{variables[i].lower()}_{normx}_train_{area}' for i,col in enumerate(columns)})
                train_columns = {col: f'{variables[i].lower()}_{normx}_train_{area}' for i,col in enumerate(columns)}
                test_columns = {col: f'{variables[i].lower()}_{normx}_test_{area}' for i,col in enumerate(columns)}


                df_train_predictor_shift = df_train['intpp_'+normx+'_train_'+area][1:]
                df_train_predictor = df_train['intpp_'+normx+'_train_'+area]

                df_test_predictor = df_test['intpp_' + normx + '_test_' + area]

                print("df_train.columns",df_train.columns)

                print("df_train_predictor_shift: ",df_train_predictor_shift)
                print("df_train['intpp_lag1_'+normx+'_train_'+area][1:]: ",df_train['intpp_lag1_'+normx+'_train_'+area][1:])

                df_new = pd.DataFrame({'Z': df_train_predictor_shift, 'Zt-1': df_train['intpp_lag1_'+normx+'_train_'+area][1:]})
                df_new_test = pd.DataFrame({'Z': df_test_predictor, 'Zt-1': df_test['intpp_lag1_'+normx+'_test_'+area]})
                i_var = 0
                column_list = ['A','B', 'C','D', 'E', 'F', 'G', 'H', 'I']
                for col in columns:

                    df_new[col] = df_train[train_columns[col]][1:]
                    df_new_test[col] = df_test[test_columns[col]]

                    scaler = MinMaxScaler()
                    scaled_data = scaler.fit_transform(df_new)
                    scaled_test_data = scaler.transform(df_new_test)

                    X, Y = self.create_dataset(scaled_data)
                    X_test, Y_test = self.create_dataset(scaled_test_data)

                    X_test = X_test[:n_forecast_steps]
                    Y_test = Y_test[:n_forecast_steps]

                    model = self.build_model(input_shape=(X.shape[1], X.shape[2]))
                    history = model.fit(X, Y, epochs=self.epochs, batch_size=self.batch_size, verbose=1)

                    self.plot_training_history(history, str(i_var), variables[i_var], normx, area)

                    y_pred = model.predict(X_test)
                    min_val, max_val = scaler.data_min_[0], scaler.data_max_[0]
                    y_pred_transformed = y_pred * (max_val - min_val) + min_val
                    Y_test_transformed = Y_test * (max_val - min_val) + min_val
                    print("[column_list[:column_list.index(col)+1]",column_list[:column_list.index(col)+1])
                    rmse = self.compute_rmse(Y_test_transformed, y_pred_transformed)
                    new_row = pd.DataFrame({'NORM': [normx], 'COLUMNS': [column_list[:column_list.index(col)+1]],
                                            'VAR': [variables[i_var]], 'RMSE': [rmse], 'AREA': [area], 'MODEL': 'LSTM', 'HYPERTUNE': '50,32,0.1,1,1'})
                    rmse_results = pd.concat([rmse_results, new_row], ignore_index=True)

                    self.plot_forecast(Y_test_transformed, y_pred_transformed, str(i_var), variables[i_var], normx, area, rmse)
                    i_var += 1
        rmse_results.to_csv(f'{self.output_lstm}/LSTM_RMSE.csv', index=False)

        end_time_lstm = time.time()
        time_used_lstm = end_time_lstm - start_time_lstm
        print("Time used for LSTM:", f'{time_used_lstm:.4f}', "Seconds")
        print("Time used for LSTM:", f'{time_used_lstm/60:.4f}', "Minutes")
        print("Time used for LSTM:", f'{time_used_lstm/3600:.4f}', "Hours")

